/**
 * Ono Project
 *
 * File:         SideStep.java
 * RCS:          $Id$
 * Description:  SideStep class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Sep 26, 2007 at 3:40:50 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.ui
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2007, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.swarmscreen.ui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.Paint;
import java.awt.event.KeyEvent;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextPane;
import javax.swing.table.DefaultTableModel;

import org.gudy.azureus2.plugins.utils.UTTimer;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;
import org.jfree.chart.ChartColor;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.DefaultDrawingSupplier;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.Range;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import edu.northwestern.swarmscreen.Main;
import edu.northwestern.swarmscreen.timer.ITimer;
import edu.northwestern.swarmscreen.timer.ITimerEvent;
import edu.northwestern.swarmscreen.timer.ITimerEventPerformer;
import edu.northwestern.swarmscreen.util.Pair;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The SideStep class is the main visualizer class for SideStep.
 */
public class SwarmScreenGraph implements UTTimerEventPerformer {

	private static final double MAX_RANGE = 200;
	private static final double MAX_Y = -1;
	private static final int MAX_HISTORY = (int)(MAX_RANGE/5);
	private static SwarmScreenGraph self;
	JFrame frame;
	private JPanel contentPane;
	private JTabbedPane tabbedPane;
	private HashMap<String, XYSeries> connRatioSeries;
	private HashMap<String, XYSeries> numConnsSeries;
	private HashMap<String, LinkedList<Pair<Long, Double>>> connRatioDataPoints;
	private HashMap<String, LinkedList<Pair<Long, Double>>> numConnsDataPoints;
	private HashMap<String, LinkedList<Pair<Long, Double>>> numTorrentsDataPoints;
	private HashMap<String, Long> lastDlTimes;
	private HashMap<String, Long> lastUlTimes;	
	private long startTime=-1;
	JFreeChart numConnsChart;
	JFreeChart connRatioChart;
	XYSeriesCollection connRatioDataset;
	XYSeriesCollection numConnsDataset;
	private JPanel graphConnRatioTab;
	private HashMap<String, String> ipToNames;
	private JPanel graphNumConnsTab;
	private JTable nearbyPeersTable;
	private DefaultTableModel nearbyPeersModel;
	private boolean ulFixedRange;
	private boolean dlFixedRange;
	LinkedList<Double> numConnsMax = new LinkedList<Double>();
	LinkedList<Double> numTorrentsMax = new LinkedList<Double>();
	LinkedList<Double> connRatioMax = new LinkedList<Double>();
	LinkedList<String> raceText = new LinkedList<String>();
	JTextPane textPane;
	private boolean active = true;
	private XYSeriesCollection numTorrentsDataset;
	private HashMap<String, XYSeries> numTorrentsSeries;
	private JFreeChart numTorrentsChart;
	private JPanel graphNumTorrentsTab;
	
	public SwarmScreenGraph(){
		connRatioDataPoints = new HashMap<String, LinkedList<Pair<Long, Double>>>();
		numConnsDataPoints = new HashMap<String, LinkedList<Pair<Long, Double>>>();
		numTorrentsDataPoints = new HashMap<String, LinkedList<Pair<Long, Double>>>();
		lastDlTimes = new HashMap<String, Long>();
		lastUlTimes = new HashMap<String, Long>();
		connRatioSeries = new HashMap<String, XYSeries>();
		numConnsSeries = new HashMap<String, XYSeries>();
		numTorrentsSeries = new HashMap<String, XYSeries>();
		ipToNames = new HashMap<String, String>();
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable()
				{
			public void run()
			{
				createAndShowGUI();
			}
				});

		synchronized(this)
		{
			while(frame==null)
			{
				try
				{
					wait();
				}
				catch(InterruptedException e)
				{
				}
			}
		}
		frame.invalidate();
		UTTimer timer = Main.getInstance().getTimer("VisualizerUpdater");
		timer.addPeriodicEvent(1000, this);
		
	}
	
	public void resetGraph(){
		connRatioSeries.clear();
		numConnsSeries.clear();
		connRatioDataset = null;
		numConnsDataset = null;
		graphConnRatioTab.removeAll();
		graphNumConnsTab.removeAll();
		numConnsChart = null;
		connRatioChart=null;
		dlFixedRange = false;
		ulFixedRange = false;
		startTime = -1;
		connRatioDataPoints.clear();
		numConnsDataPoints.clear();
		lastDlTimes.clear();
		lastUlTimes.clear();
		textPane.setText("");
		connRatioMax.clear();
		numConnsMax.clear();
		
	}
	
	
	
	public void addUlDataPoint(String ipAddress, long currentTime, double bytesSent){
		//if (bytesSent < 2000) return;
		if (numConnsDataPoints.get(ipAddress)==null) numConnsDataPoints.put(ipAddress, new LinkedList<Pair<Long, Double>>());
		synchronized (numConnsDataPoints.get(ipAddress)){
			numConnsDataPoints.get(ipAddress).add(new Pair<Long, Double>(currentTime, bytesSent));
		}
	}
	
	private synchronized void createAndShowGUI(){
		if (this.frame!=null){
			frame.setVisible(true);
		}
		JFrame frame = new JFrame("SwarmScreen Graph");
        Dimension d = java.awt.Toolkit.getDefaultToolkit().getScreenSize();

//        if (d.width < fieldX || d.height < fieldY)
//        {
//            zoom = Math.max((int)Math.ceil((double)fieldX/d.width), fieldY/(d.height-BOTTOM_OFFSET-totalTopHeight-20));
//        }
        
        //fieldX = Math.max((int)(4*infoPaneWidth)+30, fieldX);
//        d.setSize(Math.min(d.width, fieldX+10), 
//                Math.min(d.height-BOTTOM_OFFSET, fieldY/(zoom)+totalTopHeight+60));

        
		// ensure proper width for info panes

		//frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
//		frame.setBounds(0,0, this.fieldX, this.fieldY+infoPaneHeight+10);
		frame.setSize(600, 400);
        //frame.setPreferredSize(d);
		//frame.setMaximizedBounds(new Rectangle(d.width, d.height));
		
		contentPane = new JPanel();
        contentPane.setOpaque(true);
//		contentPane.setBounds(0,0, this.fieldX, this.fieldY+infoPaneHeight+10);
		//contentPane.setSize(d);
		//contentPane.setPreferredSize(d);
		
		addElementsToContentPane();
		
		frame.setContentPane(contentPane);
		
		// show frame
		frame.pack();
		frame.setVisible(true);
		
		this.frame = frame;
//        frame.setAlwaysOnTop(alwaysOnTop);
		notifyAll();
	}
	private void addElementsToContentPane() {
		tabbedPane = new JTabbedPane();
		tabbedPane.setSize(600, 400);
		tabbedPane.setPreferredSize(new Dimension(600, 400));

		graphConnRatioTab = new JPanel();
		tabbedPane.addTab("Connection ratio", null, graphConnRatioTab,
		                  "Shows current connection ratios");
		
		tabbedPane.setMnemonicAt(0, KeyEvent.VK_1);
		
		graphNumConnsTab = new JPanel();
		tabbedPane.addTab("Connections", null, graphNumConnsTab,
		                  "Shows number of connections");
		
		tabbedPane.setMnemonicAt(1, KeyEvent.VK_2);
		//addChart(graphTab);
		
		
		graphNumTorrentsTab = new JPanel();
		tabbedPane.addTab("Torrents", null, graphNumTorrentsTab,
		                  "Shows number of active torrents");		
		tabbedPane.setMnemonicAt(2, KeyEvent.VK_2);

		/*
		JPanel ssInfoTab = new JPanel(new GridBagLayout());
		tabbedPane.addTab("SS Info", null, ssInfoTab,
		                  "Information used to locate SideStep detour points");
		tabbedPane.setMnemonicAt(3, KeyEvent.VK_4);
		addContentToInfoTab(ssInfoTab);

		JPanel resultsTab = new JPanel();
		tabbedPane.addTab("Race Results", null, resultsTab,
		"Displays results from evaluating SideStep detour paths");
		tabbedPane.setMnemonicAt(4, KeyEvent.VK_5);
		textPane = new JTextPane();
		JScrollPane jsp = new JScrollPane(textPane);
		resultsTab.add(jsp);
		jsp.setSize(400, 400);
		jsp.setPreferredSize(new Dimension(400, 400));
		textPane.setAutoscrolls(true);
		jsp.setAutoscrolls(true);
		*/
		
		contentPane.add(tabbedPane);
		
	}
	
    
    private void addContentToInfoTab(JPanel ssInfoTab) {
    	
   nearbyPeersModel = new DefaultTableModel();
    
    
        String[] columnNames = {"IP",
                "Closeness",
                "Names",
                "Download",
                "DL Impr", "UL Improvement"};

		
		nearbyPeersTable = new JTable(nearbyPeersModel);
		nearbyPeersTable.setPreferredScrollableViewportSize(new Dimension(500, 70));
		nearbyPeersModel.setColumnIdentifiers(columnNames);
		//nearbyPeersTable..setFillsViewportHeight(true);
		
		
		
		//Create the scroll pane and add the table to it.
		JScrollPane scrollPane = new JScrollPane(nearbyPeersTable);
		
		//Add the scroll pane to this panel.
		ssInfoTab.add(scrollPane);
		
	}
    
  
	/**
     * Creates a chart.
     * 
     * @param dataset  the data for the chart.
     * 
     * @return a chart.
     */
    private JFreeChart createChart(final XYDataset dataset, String title, String yAxis) {
        
        // create the chart...
        JFreeChart chart = ChartFactory.createXYLineChart(
            title,      // chart title
            "Time",                      // x axis label
            yAxis,                      // y axis label
            dataset,                  // data
            PlotOrientation.VERTICAL,
            true,                     // include legend
            true,                     // tooltips
            false                     // urls
        );

        // NOW DO SOME OPTIONAL CUSTOMISATION OF THE CHART...
        chart.setBackgroundPaint(Color.white);

//        final StandardLegend legend = (StandardLegend) chart.getLegend();
  //      legend.setDisplaySeriesShapes(true);
        
        // get a reference to the plot for further customisation...
        final XYPlot plot = chart.getXYPlot();
        plot.setBackgroundPaint(new Color(220, 220, 220));
    //    plot.setAxisOffset(new Spacer(Spacer.ABSOLUTE, 5.0, 5.0, 5.0, 5.0));
        plot.setDomainGridlinePaint(Color.white);
        plot.setRangeGridlinePaint(Color.white);
        
        final XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer();
//        renderer.setSeriesFillPaint(0, Color.black);//.setSeriesLinesVisible(0, false);
//        renderer.setSeriesShapesVisible(1, false);
        renderer.setSeriesStroke(0,
        		new BasicStroke(4f, BasicStroke.JOIN_ROUND, BasicStroke.JOIN_BEVEL)                
            );
        renderer.setSeriesStroke(1,
        		new BasicStroke(4f, BasicStroke.JOIN_ROUND, BasicStroke.JOIN_BEVEL)                
            );
        plot.setRenderer(renderer);

        plot.setDrawingSupplier(getSupplier());
    
        
        // change the auto tick unit selection to integer units only...
        //final NumberAxis rangeAxis = (NumberAxis) plot.getRangeAxis();
        //rangeAxis.setStandardTickUnits(NumberAxis.createIntegerTickUnits());
        
        
        //plot.getDomainAxis().setFixedAutoRange(MAX_RANGE);
        // OPTIONAL CUSTOMISATION COMPLETED.
                
        return chart;
        
    }

	public void perform(UTTimerEvent event) {
		if (!active){
			event.cancel();
			return;
		}
		
		int WINDOW_SIZE = 5;
		// check for ip-to-name conversions
		checkForNames();
		
		// process new data points		
		if (connRatioDataset==null){
			 connRatioDataset = new XYSeriesCollection();
		}
		int numAdded = processDataPointsRatio(connRatioDataPoints, connRatioSeries, 
				connRatioDataset);
			
		
		if (connRatioChart==null && numAdded > 0){
			connRatioChart = createChart(connRatioDataset, 
					"SwarmScreen Connection Ratios", "Ratio (random:user)");
			final ChartPanel chartPanel = new ChartPanel(connRatioChart);
			
			tabbedPane.setPreferredSize(null);
	        //chartPanel.setPreferredSize(new Dimension(500, 270));
			graphConnRatioTab.add(chartPanel);
			frame.pack();
		}
		XYPlot plot;
		if (connRatioChart!=null){
			plot = connRatioChart.getXYPlot();
//			if (plot!=null) plot.getRangeAxis().setFixedAutoRange(2);
			if (plot!=null && connRatioMax.size()>0) {
				double max = Collections.max(connRatioMax);
				double min = Collections.min(connRatioMax);
				if (max>10) max = 10;
				plot.getRangeAxis().setFixedAutoRange(max);
				if (Math.abs(max-1)>1) min = 0; 
				else if (max < 2 && min > 1) min = 0.9;
				plot.getRangeAxis().setRange(new Range(min*0.9, max*1.1));
				if (plot.getDomainAxis().getRange().getUpperBound()>MAX_RANGE){
					plot.getDomainAxis().setFixedAutoRange(MAX_RANGE);
				}
			}
//			if (plot !=null && !dlFixedRange && plot.getDomainAxis().getRange().getUpperBound()>MAX_RANGE){
//				
//				
//				dlFixedRange = true;
//				plot.getDomainAxis().setFixedAutoRange(MAX_RANGE);
////				plot.getRangeAxis().setFixedAutoRange(2);
//				//plot.getRangeAxis().setUpperBound(2);
//				//if (MAX_Y!=-1) plot.getRangeAxis().setUpperBound(MAX_Y);
//			}
////			if (plot!=null && dlMax.size()>0){
////				plot.getRangeAxis().setFixedAutoRange(dlMax.getLast());
////				plot.getRangeAxis().setRange(new Range(0, dlMax.getLast()*1.1));
////
////			}

		}
		
		if (numConnsDataset==null){
			 numConnsDataset = new XYSeriesCollection();
		}
		numAdded = processDataPointsRatio(numConnsDataPoints, numConnsSeries, 
				numConnsDataset);
			
		
		if (numConnsChart==null && numAdded > 0){
			numConnsChart = createChart(numConnsDataset, "Connections", "Connections");
			final ChartPanel chartPanel = new ChartPanel(numConnsChart);
			
			tabbedPane.setPreferredSize(null);
	        //chartPanel.setPreferredSize(new Dimension(500, 270));
			graphNumConnsTab.add(chartPanel);
			frame.pack();
		}
		
		if (numConnsChart!=null){
			plot = numConnsChart.getXYPlot();
			if (plot!=null && !ulFixedRange && plot.getDomainAxis().getRange().getUpperBound()>MAX_RANGE){
				ulFixedRange = true;
				plot.getDomainAxis().setFixedAutoRange(MAX_RANGE);
			}
			if (plot!=null && numConnsMax.size()>0){
				double max = Collections.max(numConnsMax);
				plot.getRangeAxis().setFixedAutoRange(max);
				plot.getRangeAxis().setRange(new Range(0, max*1.1));

			}
		}
		
		if (numTorrentsDataset==null){
			numTorrentsDataset = new XYSeriesCollection();
		}
		numAdded = processDataPointsRatio(numTorrentsDataPoints, numTorrentsSeries, 
				numTorrentsDataset);
			
		
		if (numTorrentsChart==null && numAdded > 0){
			numTorrentsChart = createChart(numTorrentsDataset, "Torrents", "Active Torrents");
			final ChartPanel chartPanel = new ChartPanel(numTorrentsChart);
			
			tabbedPane.setPreferredSize(null);
	        //chartPanel.setPreferredSize(new Dimension(500, 270));
			graphNumTorrentsTab.add(chartPanel);
			frame.pack();
		}
		
		if (numTorrentsChart!=null){
			plot = numTorrentsChart.getXYPlot();
			if (plot!=null && plot.getDomainAxis().getRange().getUpperBound()>MAX_RANGE){
				
				plot.getDomainAxis().setFixedAutoRange(MAX_RANGE);
			}
			if (plot!=null && numTorrentsMax.size()>0){
				double max = Collections.max(numTorrentsMax);
				plot.getRangeAxis().setFixedAutoRange(max);
				plot.getRangeAxis().setRange(new Range(0, max*1.1));

			}
		}
		
		if (connRatioChart!=null){			
			connRatioChart.fireChartChanged();
		}
		if (numConnsChart!=null){
			numConnsChart.fireChartChanged();
		}
		if (numTorrentsChart!=null){
			numTorrentsChart.fireChartChanged();
		}
		
		
		LinkedList<String> temp = new LinkedList<String>();
		
		synchronized (raceText){
			temp.addAll(raceText);
		}
		for (String s : temp) textPane.setText(textPane.getText()+"\n"+s);
		synchronized (raceText){raceText.clear();	}			
	}

	/**
	 * @param windowSize
	 * @param dataset 
	 * @return
	 */
	private int processDataPoints(int windowSize, HashMap<String, LinkedList<Pair<Long, Double>>> data, 
			HashMap<String, XYSeries> series, XYSeriesCollection dataset, HashMap<String, Long> lastTimes, 
			LinkedList<Double> history) {
		int numAdded = 0;
		if (data==null) return numAdded;		
		HashSet<String> set = new HashSet<String>();
		set.addAll(data.keySet());
		long gapStart = -1;
		long gapEnd = -1;
		for (String s : set){
			LinkedList<Pair<Long, Double>> temp = new LinkedList<Pair<Long, Double>>();
			LinkedList<Pair<Long, Double>> toRemove = new LinkedList<Pair<Long, Double>>();
			if (data!=null && data.get(s)!=null){
			synchronized (data.get(s)){temp.addAll(data.get(s));}
			}
			if (temp.size()>windowSize){
				if (lastTimes.get(s)==null) {
					lastTimes.put(s, temp.getFirst().getKey());
					if (startTime==-1) startTime = temp.getFirst().getKey();
					synchronized(data.get(s)) {data.get(s).removeFirst();}
					continue;
				}
				Iterator<Pair<Long, Double>> it = temp.iterator();
				Pair<Long, Double> vals[] = new Pair[windowSize];
				for (int i = 1; i < windowSize; i++){
					vals[i] = it.next();
				}
				boolean gap = false;
				while (it.hasNext()){
					for (int i = 0; i < windowSize-1; i++){
						vals[i] = vals[i+1];
						if ( vals[i].getKey()-vals[i+1].getKey()>0){
							gapStart = vals[i].getKey();
							gapEnd = vals[i].getKey();
							gap = true;
							break;
						}
					}
					vals[windowSize-1] = it.next();
					
					if (series.get(s)==null){
						addSeriesToGraph(series, dataset, s);
					}
					long timeSum=0;
					double dataSum=0;
					for (int i = 0; i < vals.length; i++){
						timeSum+= vals[i].getKey();
						dataSum+=vals[i].getValue().doubleValue();
					}
					long timeDiff = (vals[windowSize-1].getKey()-lastTimes.get(s));					
					Double x = ((timeSum/windowSize)-startTime)/1000.0;
					Double y = (dataSum*1000.0)/timeDiff;
					if (x.isInfinite() || x.isNaN() || y.isNaN() || y.isInfinite()) continue;
					numAdded++;

					series.get(s).add(x, y, false);					
					lastTimes.put(s, vals[0].getKey());
					toRemove.add(vals[0]);
					
					if (history.size()==0){
						history.add(y);
						history.add(y);
					} else {
						if (history.size()==100){
							Double out = history.removeFirst();
							Double max = history.removeLast();
							if (out.doubleValue()==max.doubleValue()){
								max = y;
								for (Double d : history) if (d>max) max = d;								
							}		
							history.addLast(y);
							history.addLast(max);
						} else {
							Double max = history.removeLast();
							history.addLast(y);
							if (y>max) max = y;
							history.addLast(max);
						}
					}
					
				}
				
				if (gap){
					if (series.get(s).getItemCount()>0){
						series.get(s).add((startTime-gapStart+100)/1000.0, 0, false);
						series.get(s).add((startTime-gapEnd-100)/1000.0, 0, false);
						synchronized (data.get(s)){data.get(s).clear();}
					}
				}
				else {
					if (data!=null && data.get(s)!=null){
						synchronized (data.get(s)){
							data.get(s).removeAll(toRemove);
						}
					}
				}
				series.get(s).fireSeriesChanged();
			}
		}
		
		return numAdded;
	}
	
	/**
	 * @param windowSize
	 * @param dataset 
	 * @return
	 */
	private int processDataPointsRatio(HashMap<String, LinkedList<Pair<Long, Double>>> data, 
			HashMap<String, XYSeries> series, XYSeriesCollection dataset) {
		int numAdded = 0;
		if (data==null) return numAdded;		
		HashSet<String> set = new HashSet<String>();
		set.addAll(data.keySet());
		long gapStart = -1;
		long gapEnd = -1;
		for (String s : set){
			LinkedList<Pair<Long, Double>> temp = new LinkedList<Pair<Long, Double>>();
			LinkedList<Pair<Long, Double>> toRemove = new LinkedList<Pair<Long, Double>>();
			if (data!=null && data.get(s)!=null){
			synchronized (data.get(s)){temp.addAll(data.get(s));}
			}
			if (temp.size()>0){
				
				Iterator<Pair<Long, Double>> it = temp.iterator();
				
					
					if (series.get(s)==null){
						addSeriesToGraph(series, dataset, s);
					}
					while (it.hasNext()){
					Pair<Long, Double> p = it.next();
					
					Double y = p.getValue();
					if (y < 0 || y.isNaN() || y.isInfinite()){
						toRemove.add(p);
						continue;
					}
					if (startTime==-1) startTime = temp.getFirst().getKey();
					Double x = (p.getKey()-startTime)/1000.0;
					numAdded++;

					series.get(s).add(x, y, false);					
					
					toRemove.add(p);
					}
			
					if (data!=null && data.get(s)!=null){
						synchronized (data.get(s)){
							data.get(s).removeAll(toRemove);
						}
					}
				
				series.get(s).fireSeriesChanged();
			}
		}
		
		return numAdded;
	}

	/**
	 * @param series
	 * @param dataset
	 * @param s
	 */
	private void addSeriesToGraph(HashMap<String, XYSeries> series,
			XYSeriesCollection dataset, String s) {
		String name = s;
		if (ipToNames.get(s)!=null) name=ipToNames.get(s);
		series.put(s, new XYSeries(name));
		
		dataset.addSeries(series.get(s));
	}

	private void checkForNames() {
		
		HashSet<String> set = new HashSet<String>();
		set.addAll(connRatioDataPoints.keySet());
		
		lookupNames(set, connRatioSeries);
		
		set.clear();
		set.addAll(numConnsDataPoints.keySet());
		
		lookupNames(set, numConnsSeries);
		
	}

	private void lookupNames(HashSet<String> set, HashMap<String, XYSeries> series) {
		for (String n : set){
			if (ipToNames.get(n)==null){
				String name;
				try {
					 name = InetAddress.getByName(n).getHostName();
				} catch (UnknownHostException e) {
					name = n;
				}
				ipToNames.put(n, name);
				if (series.get(n)!=null) series.get(n).setDescription(name);
			}
		}
	}

	public static SwarmScreenGraph getInstance() {
		if (self == null){
			self = new SwarmScreenGraph();
		}
		return self;
		
	}
	
	 public static DefaultDrawingSupplier getSupplier(){
	        return new DefaultDrawingSupplier(
	                new Paint[] {
	                	Color.black,
	                    Color.green,
	                    Color.red,
	                    Color.orange,
	                    Color.magenta,
	                    Color.cyan,
	                    Color.pink,
	                    Color.gray,
	                    Color.yellow,
	                    ChartColor.DARK_RED,
	                    ChartColor.DARK_GREEN,
	                    ChartColor.LIGHT_RED,
	                    ChartColor.DARK_YELLOW,
	                    ChartColor.DARK_MAGENTA,
	                    ChartColor.DARK_CYAN,
	                    Color.lightGray,
	                    ChartColor.LIGHT_RED,
	                    ChartColor.LIGHT_BLUE,
	                    ChartColor.LIGHT_GREEN,
	                    ChartColor.LIGHT_YELLOW,
	                    ChartColor.LIGHT_MAGENTA,
	                    ChartColor.LIGHT_CYAN},
	                DefaultDrawingSupplier.DEFAULT_OUTLINE_PAINT_SEQUENCE,
	                            DefaultDrawingSupplier.DEFAULT_STROKE_SEQUENCE,
	                            DefaultDrawingSupplier.DEFAULT_OUTLINE_STROKE_SEQUENCE,
	                            DefaultDrawingSupplier.DEFAULT_SHAPE_SEQUENCE
	                            );
	 }
	
	public void addRaceResult(String s){
		
		//synchronized (raceText){
		raceText.add(s);
		//}
	}

	public void addConnectionTimeRatio(long time,
			double ratio) {

		String name = "Connection time ratio";
		
		if (connRatioDataPoints.get(name)==null){
//			if (dlSeries.get(ipAddress)==null){
//				addSeriesToGraph(dlSeries, dlDataset, ipAddress);
//			}
//			dlSeries.get(ipAddress).add((currentTime-startTime)/1000.0, 0);
			connRatioDataPoints.put(name, new LinkedList<Pair<Long, Double>>());
		}
		LinkedList<Pair<Long, Double>> list = connRatioDataPoints.get(name);
		synchronized (list){
//			if (list.size()>0 && list.get(list.size()-1).getValue()==0){
//				list.add(new Pair<Long, Integer>( list.get(list.size()-1).getKey()-1 ,0));
//			}

//			if (bytesSent==0){
//				if (dlSeries.get(ipAddress)==null){
//					addSeriesToGraph(dlSeries, dlDataset, ipAddress);
//				}
//				dlSeries.get(ipAddress).add((currentTime-startTime)/1000.0, 0);
//				list.clear();
//				dlDataPoints.remove(ipAddress);
//			}
//			else 
				list.add(new Pair<Long, Double>(time, ratio));
//				System.out.println((time-startTime)/1000+ " : "+connRatioMax.size());
				if (ratio>0){
					connRatioMax.addLast(ratio);
					if (connRatioMax.size()>MAX_HISTORY)
						connRatioMax.removeFirst();
				}
		}
		
		// same for target value
		if (ratio<0) return;
		
		name = "Target";
		if (connRatioDataPoints.get(name)==null){
			connRatioDataPoints.put(name, new LinkedList<Pair<Long, Double>>());
		}
		list = connRatioDataPoints.get(name);
		synchronized (list){
			list.add(new Pair<Long, Double>(time, 1.0));
		}
		
	}
	
	
	public void addConnectionCount(String type, long time,
			int count) {

		String name = type;
		
		if (numConnsDataPoints.get(name)==null){
			numConnsDataPoints.put(name, new LinkedList<Pair<Long, Double>>());
		}
		LinkedList<Pair<Long, Double>> list = numConnsDataPoints.get(name);
		synchronized (list){
				list.add(new Pair<Long, Double>(time, (double)count));
				numConnsMax.addLast((double)count);
				if (numConnsMax.size() > MAX_HISTORY*2) numConnsMax.removeFirst();
		}
		
	}
	

	public void addActiveTorrentCount(long time, int transientCount,
			int userCount) {
		
		String name = "Random";
		if (numTorrentsDataPoints.get(name)==null){
			numTorrentsDataPoints.put(name, new LinkedList<Pair<Long, Double>>());
		}
		LinkedList<Pair<Long, Double>> list = numTorrentsDataPoints.get(name);
		synchronized (list){
				list.add(new Pair<Long, Double>(time, (double)transientCount));
				numTorrentsMax.addLast((double)transientCount);
				if (numTorrentsMax.size() > MAX_HISTORY*2) numTorrentsMax.removeFirst();
		}
		
		name = "User";
		if (numTorrentsDataPoints.get(name)==null){
			numTorrentsDataPoints.put(name, new LinkedList<Pair<Long, Double>>());
		}
		list = numTorrentsDataPoints.get(name);
		synchronized (list){
				list.add(new Pair<Long, Double>(time, (double)userCount));
				numTorrentsMax.addLast((double)userCount);
				if (numTorrentsMax.size() > MAX_HISTORY*2) numTorrentsMax.removeFirst();
		}
	}

	public void close() {
		active  = false;
		frame.dispose();
		frame = null;
		
	}




}
